<?php

namespace Admin\Core\Migrations\Concerns;

use AdminCore;
use Illuminate\Support\Facades\DB;

trait SupportRelations
{
    /**
     * Check or add static relationship columns from relation admin models.
     * @param object $table
     * @param object $model
     * @param boolena $updating
     */
    public function addRelationships($table, $model, $updating = false)
    {
        $belongsToModel = $model->getBelongsToRelation();

        //Model without belongsToModel parent
        if (($count = count($belongsToModel)) == 0) {
            return;
        }

        //If update migration, then go in reverse order
        if ($updating === true) {
            $belongsToModel = array_reverse($belongsToModel);
        }

        foreach ($belongsToModel as $parent) {
            $isRecursive = class_basename(get_class($model)) == class_basename($parent);

            //If is recursive model, then do not create new same instance, because of bug when parent model
            //is overidded and has relationship on itself in package with other namespace
            $parent = $isRecursive ? $model : new $parent;

            $foreignColumn = $model->getForeignColumn($parent->getTable());

            $column = $table->integer($foreignColumn)->unsigned();

            //If parent belongs to more models, or just itself
            if (
                count($belongsToModel) > 1
                || $model->getProperty('withoutParent') === true
                || $model->getProperty('nullableRelation') === true
                || $isRecursive
            ) {
                $column->nullable();
            }

            //If foreign key does not exists in table
            if (! $model->getSchema()->hasColumn($model->getTable(), $foreignColumn)) {
                //If column does not exists in already created table, then create it after id
                if ($updating === true) {
                    $column->after('id');

                    //If is one foreign column, this columns is not null
                    //so if some rows exists, we need push values into this row
                    if (count($belongsToModel) == 1 && $model->count() > 0) {
                        $this->checkForReferenceTable($model, $foreignColumn, $parent->getTable());
                    }

                    $this->getCommand()->line('<comment>+ Added column:</comment> '.$foreignColumn);
                }
            } elseif ($updating === true) {
                $column->change();
                continue;
            }

            if ($parent->getConnection() != $model->getConnection()) {
                $this->getCommand()->line('<comment>+ Skipped foreign relationship:</comment> '.$foreignColumn.' <comment>( different db connections )</comment> ');
                continue;
            }

            $this->registerAfterAllMigrations($model, function ($table) use ($model, $foreignColumn, $parent) {
                $foreignName = $this->makeShortForeignIndex($model->getTable(), $foreignColumn, 'fkbm_');

                $table->foreign($foreignColumn, $foreignName)->references('id')->on($parent->getTable());
            });
        }
    }

    /**
     * Checks if table has already inserted rows which won't allow insert foreign key without NULL value.
     * @param  object $model
     * @param  string $key
     * @param  string $referenceTable
     * @return void
     */
    protected function checkForReferenceTable($model, $key, $referenceTable)
    {
        $this->getCommand()->line('<comment>+ Cannot add foreign key for</comment> <error>'.$key.'</error> <comment>column into</comment> <error>'.$model->getTable().'</error> <comment>table with reference on</comment> <error>'.$referenceTable.'</error> <comment>table.</comment>');
        $this->getCommand()->line('<comment>  Because table has already inserted rows. But you can insert value for existing rows for this</comment> <error>'.$key.'</error> <comment>column.</comment>');

        $referenceTableIds = AdminCore::getModelByTable($referenceTable)->take(10)->select('id')->pluck('id');

        if (count($referenceTableIds) > 0) {
            $this->getCommand()->line('<comment>+ Here are some ids from '.$referenceTable.' table:</comment> '.implode(', ', $referenceTableIds->toArray()));

            //Define ids for existing rows
            do {
                $requestedId = $this->getCommand()->ask('Which id would you like define for existing rows?');

                if (! is_numeric($requestedId)) {
                    continue;
                }

                if (AdminCore::getModelByTable($referenceTable)->where('id', $requestedId)->count() == 0) {
                    $this->getCommand()->line('<error>Id #'.$requestedId.' does not exists.</error>');
                    $requestedId = false;
                }
            } while (! is_numeric($requestedId));

            //Register event after this migration will be done
            $this->registerAfterMigration($model, function () use ($model, $key, $requestedId) {
                DB::connection($model->getConnectionName())->table($model->getTable())->update([$key => $requestedId]);
            });
        } else {
            $this->getCommand()->line('<error>+ You have to insert at least one row into '.$referenceTable.' reference table or remove all existing data in actual '.$model->getTable().' table:</error>');
            die;
        }
    }

    /**
     * Remove every second char from given string
     *
     * @param  string  $string
     * @return string
     */
    private function removeEverySecondCharInMiddle($string)
    {
        $string = str_replace('_', '', $string);

        //Split string into array
        $allChars = str_split($string);

        //Skip first and last characted in the string
        $chars = array_slice($allChars, 1, -1);

        //Remove every other characted from the middle string
        foreach ($chars as $key => $char) {
            if ( $key%2 == 0 ) {
                unset($chars[$key]);
            }
        }

        //Does not delete first and last character from the table.
        //Everything odd characted in the middle can be removed
        $newString = $allChars[0].implode('', $chars).$allChars[strlen($string)-1];

        //Return smaller character
        return strlen($newString) < strlen($string)
                ? $newString
                : $string;
    }

    /**
     * Create foreign key index name.
     * @param  string $table
     * @param  string $key
     * @return string
     */
    private function makeShortForeignIndex($table, $key, $prefix = 'fk_')
    {
        $fkStringLimit = 64;

        $table = preg_replace('/_+/', '_', $table);

        //If table name is too long for MySql
        for ( $i = 0; $i < 2; $i++ )
        {
            $totalLength = strlen($prefix) + strlen($table) + 1 + strlen($key);

            if ( strlen($table) > 10 && $totalLength > $fkStringLimit ) {
                $table = $this->removeEverySecondCharInMiddle($table);
            } else {
                break;
            }
        }

        return $prefix.$table.'_'.$key;
    }
}
